-------------------------------------------------------------------------------
--! @file       CipherCore.vhd
--! @author     Ekawat (ice) Homsirikamol
--! @brief      Top-level CipherCore module for OCB
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.ALL;
use work.AEAD_pkg.all;

entity CipherCore is
    generic (
        --! Reset behavior
        G_ASYNC_RSTN    : boolean := False; --! Async active low reset
        --! Block size (bits)
        G_DBLK_SIZE     : integer := 128;   --! Data
        G_KEY_SIZE      : integer := 128;   --! Key
        G_TAG_SIZE      : integer := 128;   --! Tag
        --! The number of bits required to hold block size expressed in
        --! bytes = log2_ceil(G_DBLK_SIZE/8)
        G_LBS_BYTES     : integer := 4;
        --! Maximum supported AD/message/ciphertext length = 2^G_MAX_LEN-1
        G_MAX_LEN       : integer := SINGLE_PASS_MAX
    );
    port (
        --! Global
        clk             : in  std_logic;
        rst             : in  std_logic;
        --! PreProcessor (data)
        key             : in  std_logic_vector(G_KEY_SIZE       -1 downto 0);
        bdi             : in  std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        --! PreProcessor (controls)
        key_ready       : out std_logic;
        key_valid       : in  std_logic;
        key_update      : in  std_logic;
        decrypt         : in  std_logic;
        bdi_ready       : out std_logic;
        bdi_valid       : in  std_logic;
        bdi_type        : in  std_logic_vector(3                -1 downto 0);
        bdi_partial     : in  std_logic;
        bdi_eot         : in  std_logic;
        bdi_eoi         : in  std_logic;
        bdi_size        : in  std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        bdi_valid_bytes : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        bdi_pad_loc     : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        --! PostProcessor
        bdo             : out std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        bdo_valid       : out std_logic;
        bdo_ready       : in  std_logic;
        bdo_size        : out std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        msg_auth_done   : out std_logic;
        msg_auth_valid  : out std_logic
    );
end entity CipherCore;

architecture structure of CipherCore is    
    constant BLOCK_SIZE         : integer := 16;
    constant LOG2_LDEPTH        : integer := log2_ceil(G_MAX_LEN-7+1);

    signal aes_almost_done      : std_logic;
    signal en_key               : std_logic;
    signal en_delta             : std_logic;
    signal en_checksum          : std_logic;
    signal en_sum               : std_logic;
    signal en_aes               : std_logic;
    signal en_l                 : std_logic;
    signal init_key             : std_logic;
    signal init_done            : std_logic;
    signal wr_l_ram             : std_logic;
    signal sel_l_addr           : std_logic;
    signal sel_l_data           : std_logic;
    signal sel_aes_decrypt      : std_logic;
    signal l_addr               : std_logic_vector(LOG2_LDEPTH  -1 downto 0);
    signal sel_data             : std_logic_vector(2            -1 downto 0);
    signal sel_delta            : std_logic;        
    signal sel_xor              : std_logic_vector(2            -1 downto 0);
    signal sel_checksum         : std_logic_vector(BLOCK_SIZE   -1 downto 0);
    signal clr_delta            : std_logic;
    signal clr_checksum         : std_logic;
    signal clr_sum              : std_logic;
    signal clr_l                : std_logic;
begin    
    u_cc_dp:
    entity work.CipherCore_Datapath(structure)
    generic map (
        G_LOG2_LDEPTH       => LOG2_LDEPTH      ,
        G_BLOCK_SIZE        => BLOCK_SIZE
    )
    port map (
        clk                 => clk              ,
        rst                 => rst              ,

        --! Input Processor
        key                 => key              ,
        bdi                 => bdi              ,

        --! Controller
        aes_almost_done     => aes_almost_done  ,
        en_key              => en_key           ,
        en_delta            => en_delta         ,
        en_checksum         => en_checksum      ,
        en_sum              => en_sum           ,
        en_aes              => en_aes           ,
        en_l                => en_l             ,
        init_key            => init_key         ,
        init_done           => init_done        ,
        l_addr              => l_addr           ,
        wr_l_ram            => wr_l_ram         ,        
        sel_l_data          => sel_l_data       ,
        sel_aes_decrypt     => sel_aes_decrypt  ,        
        sel_data            => sel_data         ,
        sel_delta           => sel_delta        ,
        sel_xor             => sel_xor          ,
        sel_checksum        => sel_checksum     ,
        clr_delta           => clr_delta        ,
        clr_checksum        => clr_checksum     ,
        clr_sum             => clr_sum          ,
        clr_l               => clr_l            ,

        --! Output Processor
        bdo                 => bdo              ,
        msg_auth_valid      => msg_auth_valid
    );

    u_cc_ctrl:
    entity work.CipherCore_Control(behavior)
    generic map (
        G_LOG2_LDEPTH       => LOG2_LDEPTH      ,
        G_BLOCK_SIZE        => BLOCK_SIZE       ,
        G_LBS_BYTES         => G_LBS_BYTES
    )
    port map (
        clk                 => clk              ,
        rst                 => rst              ,

        --! Input       
        key_ready           => key_ready        ,
        key_valid           => key_valid        ,
        key_update          => key_update       ,
        decrypt             => decrypt          ,
        bdi_ready           => bdi_ready        ,
        bdi_valid           => bdi_valid        ,
        bdi_type            => bdi_type         ,
        bdi_eot             => bdi_eot          ,
        bdi_eoi             => bdi_eoi          ,
        bdi_size            => bdi_size         ,        
        bdi_valid_bytes     => bdi_valid_bytes  ,

        --! Datapath
        aes_almost_done     => aes_almost_done  ,
        en_key              => en_key           ,
        en_delta            => en_delta         ,
        en_checksum         => en_checksum      ,
        en_sum              => en_sum           ,
        en_aes              => en_aes           ,
        en_l                => en_l             ,
        init_key            => init_key         ,
        init_done           => init_done        ,
        l_addr              => l_addr           ,
        wr_l_ram            => wr_l_ram         ,
        sel_l_data          => sel_l_data       ,
        sel_aes_decrypt     => sel_aes_decrypt  ,        
        sel_data            => sel_data         ,
        sel_delta           => sel_delta        ,
        sel_xor             => sel_xor          ,
        sel_checksum        => sel_checksum     ,
        clr_delta           => clr_delta        ,
        clr_checksum        => clr_checksum     ,
        clr_sum             => clr_sum          ,
        clr_l               => clr_l            ,

        --! Output
        msg_auth_done       => msg_auth_done    ,
        bdo_ready           => bdo_ready        ,
        bdo_valid           => bdo_valid
    );
end structure;