-------------------------------------------------------------------------------
--! @file   CipherCore_Control.vhd
--! @brief  Control unit for OCB3-AES
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.AEAD_pkg.all;

entity CipherCore_Control is
    generic (
        G_BLOCK_SIZE     : integer := 16;
        G_LOG2_LDEPTH    : integer := 5;
        G_LBS_BYTES      : integer := 4
    );
    port (
        clk              : in  std_logic;
        rst              : in  std_logic;

        --! Input
        key_valid        : in  std_logic;
        key_ready        : out std_logic;
        key_update       : in  std_logic;
        bdi_valid        : in  std_logic;
        bdi_ready        : out std_logic;
        bdi_type         : in  std_logic_vector(3                -1 downto 0);
        decrypt          : in  std_logic;
        bdi_eot          : in  std_logic;
        bdi_eoi          : in  std_logic;
        bdi_size         : in  std_logic_vector(G_LBS_BYTES+1   -1 downto 0);
        bdi_valid_bytes  : in  std_logic_vector(G_BLOCK_SIZE    -1 downto 0);

        --! Datapath
        aes_almost_done  : in  std_logic;
        init_done        : in  std_logic;
        en_key           : out std_logic;
        en_delta         : out std_logic;
        en_checksum      : out std_logic;
        en_sum           : out std_logic;
        en_aes           : out std_logic;
        en_l             : out std_logic;
        init_key         : out std_logic;
        l_addr           : out std_logic_vector(G_LOG2_LDEPTH   -1 downto 0);
        wr_l_ram         : out std_logic;
        sel_l_data       : out std_logic;
        sel_aes_decrypt  : out std_logic;
        sel_checksum     : out std_logic_vector(G_BLOCK_SIZE    -1 downto 0);
        sel_data         : out std_logic_vector(2               -1 downto 0);
        sel_delta        : out std_logic;
        sel_xor          : out std_logic_vector(2               -1 downto 0);
        clr_delta        : out std_logic;
        clr_checksum     : out std_logic;
        clr_sum          : out std_logic;
        clr_l            : out std_logic;

        --! Output
        msg_auth_done    : out std_logic;
        bdo_valid        : out std_logic;
        bdo_ready        : in  std_logic
    );
end entity CipherCore_Control;

architecture behavior of CipherCore_Control is
    function ntz_lookup ( x: std_logic_vector; size: integer )
        return std_logic_vector is
            variable output : std_logic_vector(size-1 downto 0);
    begin
        output := std_logic_vector(to_unsigned(0, size));
        for i in x'low to x'high loop
            if x(i) = '0' then
                output := std_logic_vector(unsigned(output) + 1);
            else
                return output;
            end if;
        end loop;
        return output;
    end function ntz_lookup;

    constant ZEROS              : std_logic_vector(G_BLOCK_SIZE     -1 downto 0)
        := (others => '0');

    type state_type is (
        S_INIT,             S_KEY_CHECK,
        S_LD_KEY0,          S_LD_KEY1,
        S_LD_KEY2,          S_LD_KEY3,
        S_KEY_INIT,
        S_SETUP_L_INIT,     S_SETUP_L_READ,     S_SETUP_L_WRITE,
        S_WAIT_DATA,        S_WAIT_AES,
        S_KTOP_WAIT_CPL,    S_KTOP_STALL,
        S_TAG_START,        S_TAG_OUT
        );
    signal state, nstate    : state_type;

    signal clr_counter      : std_logic;
    signal en_counter       : std_logic;
    signal counter          : std_logic_vector(G_LOG2_LDEPTH    -1 downto 0);

    signal is_bdi_nempty    : std_logic;
    signal is_full          : std_logic;
    signal msg_end          : std_logic;
    signal aes_busy         : std_logic;

    signal ntz              : std_logic_vector(G_LOG2_LDEPTH    -1 downto 0);
    signal blk_counter      : std_logic_vector(2**G_LOG2_LDEPTH-2-1 downto 0);
    signal en_blk_cnt       : std_logic;
    signal clr_blk_cnt      : std_logic;

    signal en_aes_s         : std_logic;
    signal bdi_ready_s      : std_logic;
    signal bdo_valid_s      : std_logic;
    signal is_decrypt       : std_logic;
    signal npub_read_s      : std_logic;

    signal init_key_s       : std_logic;
    signal op_decrypt       : std_logic;
    signal en_delta_s       : std_logic;
    signal en_checksum_s    : std_logic;
    signal en_sum_s         : std_logic;
    signal en_l_s           : std_logic;
    signal wr_l_ram_s       : std_logic;
    signal sel_l_addr_s     : std_logic_vector(  2              -1 downto 0);
    signal sel_l_data_s     : std_logic;
    signal sel_data_s       : std_logic_vector(  2              -1 downto 0);
    signal sel_delta_s      : std_logic;
    signal sel_xor_s        : std_logic_vector(  2              -1 downto 0);
    signal sel_checksum_s   : std_logic;
    signal sel_cksum_part   : std_logic;
    signal clr_delta_s      : std_logic;
    signal clr_checksum_s   : std_logic;
    signal clr_sum_s        : std_logic;
    signal clr_l_s          : std_logic;
    signal msg_auth_done_s  : std_logic;
    signal tag_write_s      : std_logic;

    signal bdi_ad_r         : std_logic;
    signal update_flag      : std_logic;
begin
    p_reg:
    process( clk )
    begin
        if rising_edge( clk ) then
            if rst = '1' then
                counter     <= (others => '0');
                blk_counter <= std_logic_vector(
                               to_unsigned(1,blk_counter'length));
                state       <= S_INIT;
                aes_busy    <= '0';
                msg_end     <= '0';
            else
                if (clr_counter = '1') then
                    counter <= (others => '0');
                elsif (en_counter = '1') then
                    counter <= std_logic_vector(unsigned(counter) + 1);
                end if;

                state <= nstate;

                if (en_aes_s = '1') then
                    aes_busy <= '1';
                elsif (aes_almost_done = '1') then
                    aes_busy <= '0';
                end if;

                if (clr_blk_cnt = '1') then
                    blk_counter <= std_logic_vector(
                                   to_unsigned(1,blk_counter'length));
                elsif (en_blk_cnt = '1') then
                    blk_counter <= std_logic_vector(unsigned(blk_counter) + 1);
                end if;

                if (clr_sum_s = '1') then
                    bdi_ad_r    <= '0';
                    msg_end     <= '0';
                elsif (bdi_ready_s = '1' or update_flag = '1') then
                    if (bdi_type(2 downto 1) = "00") then
                        bdi_ad_r        <= '1';
                    else
                        bdi_ad_r        <= '0';
                    end if;
                    msg_end         <= bdi_eoi;
                    is_full         <= bdi_size(G_LBS_BYTES);
                end if;

                if (bdi_ready_s = '1' and bdi_type = BDI_TYPE_NPUB) then
                    is_decrypt <= decrypt;
                end if;

                --! Output registers
                en_aes          <= en_aes_s      ;
                en_delta        <= en_delta_s    ;
                en_checksum     <= en_checksum_s ;
                en_sum          <= en_sum_s      ;
                sel_aes_decrypt <= op_decrypt   ;
                sel_data        <= sel_data_s    ;
                sel_delta       <= sel_delta_s   ;
                sel_xor         <= sel_xor_s     ;
                bdi_ready       <= bdi_ready_s   ;
                bdo_valid       <= bdo_valid_s   ;
                init_key        <= init_key_s    ;
                msg_auth_done   <= msg_auth_done_s;

                sel_checksum(0)  <= sel_checksum_s or sel_cksum_part;
                for i in 1 to G_BLOCK_SIZE-1 loop
                    sel_checksum(i)  <= sel_checksum_s
                        or (sel_cksum_part
                            and bdi_valid_bytes(G_BLOCK_SIZE-1-i));
                end loop;
            end if;
        end if;
    end process;
    is_bdi_nempty   <= '1' when unsigned(bdi_size) /= 0 else '0';

    ntz            <= ntz_lookup(blk_counter, G_LOG2_LDEPTH);   --! Note: Maybe should convert this to a Fixed ROM to ensure the size
    --! Unregistered output
    with sel_l_addr_s select
    l_addr          <=  std_logic_vector(unsigned(ntz)+2)       when "00",     --! L_ntz(i) = L_ntz
                        counter                                 when "01",     --! L_* = L[0] OR counter (used for initialization)
                        (ZEROS(G_LOG2_LDEPTH-1 downto 1) & '1') when others;   --! L_$ = L[1]

    en_l            <= en_l_s        ;
    wr_l_ram        <= wr_l_ram_s    ;
    sel_l_data      <= sel_l_data_s  ;
    clr_l           <= clr_l_s       ;
    clr_delta       <= clr_delta_s   ;
    clr_checksum    <= clr_checksum_s;
    clr_sum         <= clr_sum_s     ;

    p_state:
    process( state, bdi_valid, bdi_type, is_decrypt,
             bdi_eot, bdi_eoi, aes_busy, bdo_ready, is_bdi_nempty,
             key_valid, key_update, counter, msg_end, init_done,
             aes_almost_done, bdi_ad_r, is_full)
    begin
        --! Default values
        --! ++ Externals
        npub_read_s       <= '0';
        bdi_ready_s       <= '0';
        tag_write_s       <= '0';
        msg_auth_done_s   <= '0';

        en_key            <= '0';
        en_delta_s        <= '0';
        en_checksum_s     <= '0';
        en_sum_s          <= '0';
        en_aes_s          <= '0';
        en_l_s            <= '0';
        wr_l_ram_s        <= '0';
        sel_l_addr_s      <= "00";
        sel_l_data_s      <= '0';
        sel_data_s        <= "00";
        sel_delta_s       <= '0';
        sel_xor_s         <= "00";
        clr_l_s           <= '0';
        clr_delta_s       <= '0';
        clr_checksum_s    <= '0';
        clr_sum_s         <= '0';
        op_decrypt <= '0';
        init_key_s        <= '0';
        sel_checksum_s    <= '0';
        key_ready         <= '0';
        bdo_valid_s       <= '0';
        sel_cksum_part    <= '0';
        update_flag       <= '0';

        --! ++ Internals
        nstate            <= state;
        en_blk_cnt        <= '0';
        clr_blk_cnt       <= '0';
        en_counter        <= '0';
        clr_counter       <= '0';

        case state is
            when S_INIT =>
                nstate          <= S_KEY_CHECK;
                clr_counter     <= '1';
                clr_delta_s     <= '1';
                clr_sum_s       <= '1';
                clr_checksum_s  <= '1';
                clr_l_s         <= '1';
                clr_blk_cnt     <= '1';

            when S_KEY_CHECK =>
                clr_l_s         <= '1';
                if (bdi_valid = '1' or key_update = '1') then
                    if (key_update = '1') then   --! Initialize new key
                        if (key_valid = '1') then
                            nstate      <= S_LD_KEY0;
                        end if;
                    else
                        nstate      <= S_WAIT_DATA;
                        clr_counter <= '1';
                    end if;
                end if;

            when S_LD_KEY0 =>
                key_ready <= '1';
                if (key_valid = '1') then
                    en_key <= '1';
                    nstate <= S_LD_KEY1;
                end if;
            when S_LD_KEY1 =>
                key_ready <= '1';
                if (key_valid = '1') then
                    en_key <= '1';
                    nstate <= S_LD_KEY2;
                end if;
            when S_LD_KEY2 =>
                key_ready <= '1';
                if (key_valid = '1') then
                    en_key <= '1';
                    nstate <= S_LD_KEY3;
                end if;
            when S_LD_KEY3 =>
                key_ready <= '1';
                if (key_valid = '1') then
                    en_key      <= '1';
                    nstate      <= S_KEY_INIT;
                    init_key_s  <= '1';
                end if;

            WHEN S_KEY_INIT =>
                if init_done = '1' then     --! ENCIPHER(K, zeros(128))
                    nstate      <= S_SETUP_L_INIT;
                    en_aes_s    <= '1';             --! Start AES unit
                    sel_data_s  <= "11";            --! Select 0 data
                end if;

            when S_SETUP_L_INIT =>
                if aes_busy = '0' then
                    sel_l_addr_s <= "01";
                    sel_l_data_s <= '1';
                    wr_l_ram_s   <= '1';
                    nstate       <= S_SETUP_L_READ;
                end if;

            when S_SETUP_L_READ =>
                nstate          <= S_SETUP_L_WRITE;
                sel_l_addr_s    <= "01";
                en_l_s          <= '1';
                en_counter      <= '1';

            when S_SETUP_L_WRITE =>
                sel_l_addr_s    <= "01";
                wr_l_ram_s      <= '1';

                if  (unsigned(counter) = 2**G_LOG2_LDEPTH - 1) then
                    nstate      <= S_WAIT_DATA;
                    clr_counter <= '1';
                else
                    nstate      <= S_SETUP_L_READ;
                end if;

            when S_WAIT_DATA =>
                if (bdi_valid = '1' and aes_busy = '0') then
                    update_flag     <= '1';
                    if (bdi_type = BDI_TYPE_NPUB) then
                        --! Calculate Ktop
                        nstate          <= S_KTOP_WAIT_CPL;
                        sel_data_s      <= "01";       --! Select Data = Nonce[1..122]||zeros(6)
                        clr_l_s         <= '1';        --! Clear L      (this clock cycle)
                        clr_delta_s     <= '1';        --! Clear offset (this clock cycle)
                        en_aes_s        <= '1';        --! Start AES
                    elsif (bdi_type(2 downto 1) = "00") then
                        --! Calculate Data
                        nstate          <= S_WAIT_AES;
                        if (bdi_size(G_LBS_BYTES) = '0') then
                            sel_l_addr_s <= "01";       --! Select L_*
                        end if;
                        bdi_ready_s     <= '1';
                        en_l_s          <= '1';         --! Load next L
                        en_blk_cnt      <= '1';         --! increment block counter
                        en_delta_s      <= '1';         --! Update offset
                        en_aes_s        <= '1';         --! Start AES
                    elsif (bdi_type(2 downto 1) = "01") then
                        nstate          <= S_WAIT_AES;
                        if (bdi_size(G_LBS_BYTES) = '0') then
                            --! Only apply to last block
                            sel_data_s   <= "11";    --! Select 0
                            sel_l_addr_s <= "01";    --! Select L_*
                        else
                            bdi_ready_s  <= '1';
                            if (is_decrypt = '1') then
                                op_decrypt <= '1';    --! Select decryption operation
                            else
                                en_checksum_s     <= '1';    --! Update checksum if encryption
                            end if;
                        end if;
                        en_l_s          <= '1';         --! Load next L
                        en_blk_cnt      <= '1';         --! increment block counter
                        en_delta_s      <= '1';         --! Update offset
                        en_aes_s        <= '1';         --! Start AES
                    end if;
                end if;

            when S_WAIT_AES =>
                if (aes_almost_done = '1' or aes_busy = '0') then
                    if (bdi_ad_r = '1') then
                        sel_xor_s   <= "10";
                        en_sum_s    <= '1';
                        nstate      <= S_WAIT_DATA;
                    elsif (bdo_ready = '1') then
                        if (msg_end = '1') then --! Last data block
                            nstate <= S_TAG_START;
                        else
                            nstate <= S_WAIT_DATA;
                        end if;
                        bdo_valid_s <= '1';

                        if (is_full = '1') then
                            if (is_decrypt = '1') then
                                sel_checksum_s  <= '1';    --! Select Plaintext output (AES_DATA xor OFFSET_i)
                                en_checksum_s   <= '1';    --! Update Checksum
                            end if;
                        else
                            bdi_ready_s         <= '1';    --! Read last block
                            sel_xor_s           <= "01";   --! Select data for output
                            en_checksum_s       <= '1';    --! Update Checksum
                            if (is_decrypt = '1') then
                                sel_cksum_part <= '1';     --! Select P_* for decryption
                            end if;
                        end if;
                    end if;
                end if;

            when S_KTOP_WAIT_CPL =>         --! Wait for Ktop to complete and finish the init routine for OCB-Crypt
                if (aes_almost_done = '1' and bdi_valid = '1') then
                    if (bdi_eoi = '1') then
                        nstate          <= S_TAG_START;
                    else
                        nstate          <= S_KTOP_STALL;
                    end if;
                    en_delta_s      <= '1';         --! Update offset
                    sel_delta_s     <= '1';         --! Load it with Stretch[1+bottom..128+bottom]
                    bdi_ready_s     <= '1';         --! Npub is no longer needed
                    clr_checksum_s  <= '1';         --! Checksum = 0
                    clr_blk_cnt     <= '1';         --! Block count = 1
                end if;

            when S_KTOP_STALL =>
                nstate <= S_WAIT_DATA;

            when S_TAG_START =>
                if (aes_busy = '0' or aes_almost_done = '1') then
                    nstate              <= S_TAG_OUT;
                    sel_l_addr_s        <= "10";    --! Select L_$
                    en_l_s              <= '1';     --! Load next L
                    sel_data_s          <= "10";    --! Select Checksum data
                    en_aes_s            <= '1';     --! Start AES
                end if;

            when S_TAG_OUT =>
                if ((aes_busy = '0' or aes_almost_done = '1')
                    and ((is_decrypt = '0' and bdo_ready = '1')
                        or (is_decrypt = '1' and bdi_valid = '1')))
                then
                    nstate              <= S_INIT;
                    if (is_decrypt = '0') then
                        bdo_valid_s     <= '1';     --! Write output
                    else
                        bdi_ready_s     <= '1';     --! Read tag
                        msg_auth_done_s <= '1';     --! Message auth done
                    end if;
                    sel_xor_s           <= "10";    --! Select HASH(K,A)
                end if;
        end case;
    end process;

end behavior;
