-------------------------------------------------------------------------------
--! @file   CipherCore_Datapath.vhd
--! @brief  Datapath unit for OCB
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity CipherCore_Datapath is
    generic (
        G_LOG2_LDEPTH   : integer := 5;
        G_BLOCK_SIZE    : integer := 16
    );
    port    (
        clk             : in  std_logic;
        rst             : in  std_logic;

        --! Input
        bdi             : in  std_logic_vector(128              -1 downto 0);
        key             : in  std_logic_vector(32               -1 downto 0);

        --! Datapath
        aes_almost_done : out std_logic;
        init_done       : out std_logic;
        en_key          : in  std_logic;
        en_delta        : in  std_logic;
        en_checksum     : in  std_logic;
        en_sum          : in  std_logic;
        en_aes          : in  std_logic;
        en_l            : in  std_logic;
        l_addr          : in  std_logic_vector(G_LOG2_LDEPTH    -1 downto 0);
        wr_l_ram        : in  std_logic;
        sel_l_data      : in  std_logic;
        sel_aes_decrypt : in  std_logic;
        init_key        : in  std_logic;
        sel_data        : in  std_logic_vector(2                -1 downto 0);
        sel_delta       : in  std_logic;
        sel_xor         : in  std_logic_vector(2                -1 downto 0);
        sel_checksum    : in  std_logic_vector(G_BLOCK_SIZE     -1 downto 0);
        clr_delta       : in  std_logic;
        clr_checksum    : in  std_logic;
        clr_sum         : in  std_logic;
        clr_l           : in  std_logic;

        --! Output
        bdo             : out std_logic_vector(128              -1 downto 0);
        msg_auth_valid  : out std_logic
    );
end entity CipherCore_Datapath;

architecture structure of CipherCore_Datapath is
    type ram_type is array (0 to 2**G_LOG2_LDEPTH-1) of
        std_logic_vector(128            -1 downto 0);
    signal l_ram        : ram_type;
    signal reg_l_msb    : std_logic_vector(  8                  -1 downto 0);
    signal doubled_l    : std_logic_vector(128                  -1 downto 0);


    signal top          : std_logic_vector(128                  -1 downto 0);
    signal aes_data     : std_logic_vector(128                  -1 downto 0);
    signal stretch      : std_logic_vector(256                  -1 downto 0);
    signal stretch_shf  : std_logic_vector(256                  -1 downto 0);
    signal bottom       : std_logic_vector(  6                  -1 downto 0);
    signal delta_init   : std_logic_vector(128                  -1 downto 0);

    signal mux_delta    : std_logic_vector(128                  -1 downto 0);
    signal mux_l_data   : std_logic_vector(128                  -1 downto 0);
    signal mux_data     : std_logic_vector(128                  -1 downto 0);
    signal mux_xor4     : std_logic_vector(128                  -1 downto 0);
    signal mux_checksum : std_logic_vector(128                  -1 downto 0);

    signal reg_key      : std_logic_vector(128                  -1 downto 0);
    signal reg_sum      : std_logic_vector(128                  -1 downto 0);
    signal reg_delta    : std_logic_vector(128                  -1 downto 0);
    signal reg_l        : std_logic_vector(128                  -1 downto 0);
    signal reg_checksum : std_logic_vector(128                  -1 downto 0);

    signal xor1         : std_logic_vector(128                  -1 downto 0);
    signal xor2         : std_logic_vector(128                  -1 downto 0);
    signal xor3         : std_logic_vector(128                  -1 downto 0);
    signal xor4         : std_logic_vector(128                  -1 downto 0);

    constant ZEROS      : std_logic_vector(128                  -1 downto 0)
        := (others => '0');
begin

    process(clk)
    begin
        if rising_edge(clk) then
            if (en_key = '1') then
                reg_key <= reg_key(95 downto 0) & key;
            end if;

            if (clr_sum = '1') then
                reg_sum <= (others => '0');
            elsif (en_sum = '1') then
                reg_sum <= xor4;
            end if;

            if (clr_delta = '1') then
                reg_delta <= (others => '0');
            elsif (en_delta = '1') then
                reg_delta <= mux_delta;
            end if;

            if (clr_sum = '1') then
                reg_checksum <= (others => '0');
            elsif (en_checksum = '1') then
                reg_checksum <= xor3;
            end if;
        end if;
    end process;

    uAES:
    entity work.AES_EncDec(structure)
    generic map (G_OBUF => True)
    port map (clk => clk, rst=>rst,
              din => xor2,  key=>reg_key, dout=>aes_data,
              init => init_key, done_init => init_done,
              decrypt=>sel_aes_decrypt, start=>en_aes,
              ready=>open, almost_done=>aes_almost_done );

    p_lram:
    process( clk )
    begin
        if rising_edge( clk ) then
            if wr_l_ram = '1' then
                l_ram(to_integer(unsigned(l_addr))) <= mux_l_data;
            end if;

            if clr_l = '1' then
                reg_l   <= (others => '0');
            elsif en_l = '1' then
                reg_l   <= l_ram(to_integer(unsigned(l_addr)));
            end if;
        end if;
    end process;

    reg_l_msb <= x"87" when reg_l(127) = '1' else x"00";
    doubled_l <= reg_l(126 downto 7) & ((reg_l(6 downto 0) &'0') xor reg_l_msb);

    top       <= x"0000_000" & "0001" & bdi(127 downto 38) & "000000";
    stretch   <= aes_data & (aes_data xor (aes_data(119 downto 0) & x"00"));

    u_bshift:
    entity work.bshift(struct)
    generic map ( G_W => 256, G_LOG2_W => 6, G_LEFT => 1 )
    port map    ( ii => stretch, rtr => bottom, oo => stretch_shf );

    bottom      <= bdi(37 downto 32);
    delta_init  <= stretch_shf(255 downto 128);

    xor1        <= reg_delta    xor reg_l;
    xor2        <= xor1         xor mux_data;
    xor3        <= mux_checksum xor reg_checksum;
    xor4        <= aes_data     xor mux_xor4;
    bdo         <= xor4;

    mux_delta   <= delta_init when sel_delta   = '1' else xor1;
    mux_l_data  <= aes_data   when sel_l_data  = '1' else doubled_l;

    with sel_xor select
    mux_xor4    <=  reg_delta       when "00",
                    bdi             when "01",
                    reg_sum         when others;

    msg_auth_valid <= '1' when (xor4 = bdi) else '0';

    with sel_data select
    mux_data    <=  bdi             when "00",
                    top             when "01",
                    reg_checksum    when "10",
                    (others => '0') when others;

    g_mux_checksum:
    for i in 0 to G_BLOCK_SIZE-1 generate
        mux_checksum(i*8+7 downto i*8)    <= bdi(i*8+7 downto i*8)
            when sel_checksum(G_BLOCK_SIZE-1-i) = '0'
            else xor4(i*8+7 downto i*8);
    end generate;
end architecture structure;